/*
 * ProducerConsumersQueue.hpp
 *
 *  Created on: Aug 12, 2009
 *      Author: kottlers
 */

#ifndef LOCKFREEQUEUES_HPP_
#define LOCKFREEQUEUES_HPP_
#define lockfreequeues_hpp_version "$Revision: 1.5 $"

#include "GlobalSettings.hpp"
#include "Logging.hpp"
#include "SharedBasic.hpp"

/*
template<class E>
struct DbgEle{
	E 			e;
	uint32_t 	id;
	explicit DbgEle():e(E()),id(max32){}
	explicit DbgEle(const E& ele,uint32_t i):e(ele),id(i){}
};

template<class T>
struct Conv{
	static char* toString(){return __STRING(T);}
};

class SharedClause;
class GraphUpdate;
class ShortCls;

template<>
struct Conv<SharedClause*>{
	static char* toString(){return "SharedClause";}
};

template<>
struct Conv<GraphUpdate>{
	static char* toString(){return "GraphUpdate";}
};

template<>
struct Conv<ShortCls>{
	static char* toString(){return "ShortCls";}
};

*/

template<class DTA>
struct GrowableQueueData : public UpdateChain<GrowableQueueData<DTA> >{

	typedef UpdateChain<GrowableQueueData> super;

	const uint32_t			modulo_mask;
	volatile uint32_t		w_head;
	uint32_t				r_head;

	DTA*					data;
	//DbgEle<DTA>*			data;

private:
	~GrowableQueueData(){delete [] data;
	LOG3("Free unused data in GrowableQueueData " << (modulo_mask+1)*sizeof(DTA) << " bytes");}

	/**Copying of Clause instances is not allowed*/
	GrowableQueueData(const GrowableQueueData<DTA>& cl):super(cl){asrt(false);}
	GrowableQueueData& operator=(const GrowableQueueData<DTA>& c){asrt(false);return c;}
public:
	inline GrowableQueueData(const UserMask& all_users, const uint32_t& mod_mask);

	uint32_t	getNext(const uint32_t& t){return (t + 1) & modulo_mask;}

	void		incRHead()					{r_head = getNext(r_head);}

	DTA&		operator [] (uint32_t i){asrt((i&modulo_mask)==i);return data[i];}
	//DbgEle<DTA>&		operator [] (uint32_t i){asrt((i&modulo_mask)==i);return data[i];}

	void		linkUpdate(GrowableQueueData<DTA>* updt){super::linkUpdate(updt);}

	//interface for UpdateChain
	static void destruct(GrowableQueueData<DTA>* d){delete d;}

	static GrowableQueueData<DTA>* loadUpdate(GrowableQueueData<DTA>* d,const UserBit& b){
			return super::loadNextUpdate(d,b);}
};




template<class DTA>
struct GrowableMultQueueData : public UpdateChain<GrowableMultQueueData<DTA> >{

	typedef UpdateChain<GrowableMultQueueData> super;

	const uint32_t			modulo_mask;
	volatile uint32_t		w_head;
	uint32_t*				r_heads;

	DTA*					data;
	//DbgEle<DTA>*			data;
private:
	~GrowableMultQueueData(){delete[] r_heads; delete[] data;
	LOG3("Free unused data in GrowableMultQueueData " << (modulo_mask+1)*sizeof(DTA) << " bytes");}

	/**Copying of Clause instances is not allowed*/
	GrowableMultQueueData(const GrowableMultQueueData<DTA>& cl):super(cl){asrt(false);}
	GrowableMultQueueData& operator=(const GrowableMultQueueData<DTA>& c){asrt(false);return c;}
public:
	inline GrowableMultQueueData(const UserMask& all_users, const uint32_t consumers, const uint32_t& mod_mask);

	uint32_t	getNext(const uint32_t& t){return (t + 1) & modulo_mask;}

	inline bool oneReadHeadEquals(const uint32_t& t,uint32_t cons);

	DTA&		operator [] (uint32_t i){asrt((i&modulo_mask)==i);return data[i];}
	//DbgEle<DTA>&	operator [] (uint32_t i){asrt((i&modulo_mask)==i);return data[i];}

	void		linkUpdate(GrowableMultQueueData<DTA>* updt){super::linkUpdate(updt);}

	//interface for UpdateChain
	static 		void destruct(GrowableMultQueueData<DTA>* d){delete d;}

	static 		GrowableMultQueueData<DTA>* loadUpdate(GrowableMultQueueData<DTA>* d,const UserBit& b){
				return super::loadNextUpdate(d,b);}
};




template<class T>
class GrowableOneReadQueue{
	static const uint32_t INITIAL_MODULO_MASK = (1 << 4)-1;
	static uint32_t	incModMask(const uint32_t& m){return (m<<1)+1;}
	typedef GrowableQueueData<T> data_t;
public:
	const UserID 	owner_id;
	const UserMask 	all_users;
private:
	data_t*  		read_d;
	data_t*  		write_d;

	//uint32_t				dbg_next_w;
	//uint32_t				dbg_next_r;

	GrowableOneReadQueue(const GrowableOneReadQueue& users){asrt(false);}

	inline data_t* construct(const uint32_t mod_mask){return new data_t(all_users,mod_mask);}

public:
	inline GrowableOneReadQueue(const UserMask& users, const UserID& uid);
	inline ~GrowableOneReadQueue();

	inline bool pop(const UserBit& u_bit, T& out);

	inline void push(const T& e);

	inline void push(const vector<T>& unit);

	template<class Q,class X>
	friend void pushAllocNew(Q& q,const vector<X>& unit);
};





template<class T>
class GrowableMultReadQueue{
	static const uint32_t INITIAL_MODULO_MASK = (1 << 4)-1;
	static uint32_t	incModMask(const uint32_t& m){return (m<<1)+1;}
	typedef GrowableMultQueueData<T> data_t;

	const UserID		owner_id;
	const uint16_t		consumers;
	const UserMask 		all_users;

	data_t*  			write_d;
	data_t**  			read_ds;

	//uint32_t				dbg_next_w;
	//uint32_t*				dbg_next_r;

	inline uint32_t adaptID(const uint32_t id) const;

	GrowableMultReadQueue(const GrowableMultReadQueue& users){asrt(false);}

	inline data_t* construct(const uint32_t mod_mask){return new data_t(all_users,consumers,mod_mask);}

public:
	inline GrowableMultReadQueue(const UserMask& users,const UserID& uid,const uint32_t cons);
	inline ~GrowableMultReadQueue();

	inline bool pop(const UserBit& u_bit,const UserID& uid, T& out);
	//inline bool canPop(const UserID& uid) const ;

	inline void push(const T& e);

	inline void push(const vector<T>& unit);

	template<class Q,class X>
	friend void pushAllocNew(Q& q,const vector<X>& unit);
};

//implementation

template<class T>
bool GrowableOneReadQueue<T>::pop(const UserBit& u_bit, T& out){
	asrt(read_d != NULL && write_d != NULL);

	while(read_d->r_head == read_d->w_head){
		if(!read_d->hasUpdate()) return false;

		//check again to handle bad cases where the writing thread extended read_d at one push and linked update at next push
		//and did these pushes all between the last "r_head==w_head" check and the following "hasUpdate" check
		if(read_d->r_head == read_d->w_head){
			read_d = data_t::loadUpdate(read_d,u_bit);
			LOG3("Load update data in GrowableOneReadQueue" );}
		else{continue;}

		asrt(read_d->r_head==0);
		if(read_d->w_head==0) return false;
	}
	/*DbgEle<T>& tmp = (*read_d)[read_d->r_head];
	out				=tmp.e;
	LOG_MEDIUM("GrowableOneReadQueue "<< Conv<T>::toString() <<" pop element " << tmp.id);
	asrt(dbg_next_r==tmp.id);
	dbg_next_r++;*/

	out		= (*read_d)[read_d->r_head];
	read_d	->incRHead();
	return true;
}

/*
template<class T>
bool GrowableMultReadQueue<T>::canPop(const UserID& uid) const {
	asrt(consumers > 0);
	uint32_t id 	= adaptID(uid.id); 	asrt(id < consumers);
	data_t* read_d 	= read_ds[id];		asrt(read_d != NULL && write_d != NULL);

	return read_d->r_heads[id] != read_d->w_head || read_d->hasUpdate();
}*/

template<class T>
bool GrowableMultReadQueue<T>::pop(const UserBit& u_bit,const UserID& uid, T& out){
	asrt(consumers > 0);
	uint32_t id 	= adaptID(uid.id); 	asrt(id < consumers);
	data_t* read_d 	= read_ds[id];		asrt(read_d != NULL && write_d != NULL);

	while(read_d->r_heads[id] == read_d->w_head){
		if(!read_d->hasUpdate()){return false;}

		//check again to handle bad cases where the writing thread extended read_d at one push and linked update at next push
		//and did these pushes all between the last "r_head==w_head" check and the following "hasUpdate" check
		if(read_d->r_heads[id] == read_d->w_head){
			read_d 		= data_t::loadUpdate(read_d,u_bit);	asrt(read_d != read_ds[id]);
			read_ds[id]	= read_d;
			LOG3("Load update data in GrowableMultReadQueue" );}
		else{continue;}

		asrt(read_d->r_heads[id]==0);
		if(read_d->w_head==0) return false;
	}

	const uint32_t rhd 	= read_d->r_heads[id];
	/*DbgEle<T>& tmp = (*read_d)[rhd];
	out				=tmp.e;
	LOG_MEDIUM("GrowableMultReadQueue "<< Conv<T>::toString() <<" pop element " << tmp.id);
	asrt(dbg_next_r[id]==tmp.id);
	dbg_next_r[id]++;*/
	out					= (*read_d)[rhd];
	read_d->r_heads[id]	= read_d->getNext(rhd);
	return true;
}


template<class T>
void GrowableOneReadQueue<T>::push(const T& e){
	asrt(read_d != NULL && write_d != NULL);
	data_t& w = *write_d;
	uint32_t next_w = w.getNext((const uint32_t)w.w_head);

	//LOG_MEDIUM("GrowableOneReadQueue "<< Conv<T>::toString() <<" push element " << dbg_next_w);
	if(next_w == w.r_head){

		data_t* newwd	= construct(incModMask(w.modulo_mask));
		(*newwd)[0]		= e; asrt(newwd->w_head==0);
		//(*newwd)[0]		= DbgEle<T>(e,dbg_next_w++); asrt(newwd->w_head==0);
		newwd->w_head 	= 1;

		w.linkUpdate(newwd);	asrt(w.hasUpdate());
		write_d 		= data_t::loadUpdate(write_d,UserBit(owner_id)); asrt(!write_d->hasUpdate());
	}else{
		//w[w.w_head]		= DbgEle<T>(e,dbg_next_w++);
		w[w.w_head]		= e;
		w.w_head 		= next_w;
	}
}


template<class T>
void GrowableMultReadQueue<T>::push( const T& e){
	if(consumers == 0) return;

	asrt(write_d != NULL);
	uint32_t next_w = write_d->getNext((const uint32_t)write_d->w_head);

	//LOG_MEDIUM("GrowableMultReadQueue "<< Conv<T>::toString() <<" push element " << dbg_next_w);
	if(write_d->oneReadHeadEquals(next_w,consumers)){

		data_t* newwd	= construct(incModMask(write_d->modulo_mask));
		//(*newwd)[0]		= DbgEle<T>(e,dbg_next_w++);asrt(newwd->w_head==0);
		(*newwd)[0]		= e; asrt(newwd->w_head==0);
		newwd->w_head 	= 1;

		write_d			->linkUpdate(newwd);	asrt(write_d->hasUpdate());
		write_d 		= data_t::loadUpdate(write_d,UserBit(owner_id)); asrt(!write_d->hasUpdate());
	}else{
		//(*write_d)[write_d->w_head]	= DbgEle<T>(e,dbg_next_w++);
		(*write_d)[write_d->w_head]	= e;
		write_d->w_head 			= next_w;
	}
}



template<class Q,class T>
inline void pushAllocNew(Q& queue, const vector<T>& unit){
	uint32_t mod_mask 		=	queue.write_d->modulo_mask;
	const uint32_t mm_bkup	=	mod_mask < unit.size() ? mod_mask : 0;
	while(mod_mask<unit.size()){mod_mask=queue.incModMask(mod_mask);}

	uint32_t u=0;
	typename Q::data_t* newwd = queue.construct(mod_mask); asrt(newwd->w_head==0);
	for(;u!=unit.size();++u){ (*newwd)[u]=unit[u];}
	newwd->w_head 	= u; asrt((u&mod_mask)==u);

	queue.write_d			->linkUpdate(newwd);	asrt(queue.write_d->hasUpdate());
	queue.write_d 			= Q::data_t::loadUpdate(queue.write_d,UserBit(queue.owner_id));
	asrt(!queue.write_d->hasUpdate());

	//this function is usually called when a big amount of data has to be sent.
	//Usually later on the queue does not need so much mem anymore! -->
	//Link new smaller update so that the big block of mem is freed as soon as possible
	if(mm_bkup == 0) 	return; //new memory bolck was not too big -> leave it be

	newwd 				= queue.construct(mm_bkup); asrt(newwd->w_head==0);
	queue.write_d		->linkUpdate(newwd);	asrt(queue.write_d->hasUpdate());
	queue.write_d 		= Q::data_t::loadUpdate(queue.write_d,UserBit(queue.owner_id));
}




template<class T>
void GrowableOneReadQueue<T>::push( const vector<T>& unit){
	asrt(read_d != NULL && write_d != NULL);
	//CONSOLE("Push vector to GrowableOneReadQueue with size " << unit.size() << endl);

	if(unit.empty()){ 						return;}
	if(write_d->modulo_mask<=unit.size()){	pushAllocNew(*this,unit);return;}

	uint32_t w_head = write_d->w_head;
	//const uint32_t next_w_dbg = dbg_next_w;
	for(uint32_t u=0;u!=unit.size();++u){
		uint32_t next_w 		= write_d->getNext(w_head);

		if(next_w != write_d->r_head){
			//LOG_MEDIUM("GrowableOneReadQueue "<< Conv<T>::toString() <<" push element " << dbg_next_w);
			(*write_d)[w_head]	= unit[u];
			w_head 				= next_w;
		}else{
			//dbg_next_w = next_w_dbg;
			pushAllocNew(*this,unit);	return;
		}
	}
	write_d->w_head			= w_head;
}


template<class T>
void GrowableMultReadQueue<T>::push(const vector<T>& unit){
	asrt(write_d != NULL);
	//CONSOLE("Push vector to GrowableMultReadQueue with size " << unit.size() << endl);
	if(consumers == 0 || unit.empty()){ 	return;}
	if(write_d->modulo_mask<=unit.size()){	pushAllocNew(*this,unit);return;}

	uint32_t w_head	= write_d->w_head;
	//const uint32_t next_w_dbg = dbg_next_w;
	for(uint32_t u=0;u!=unit.size();++u){
		uint32_t next_w = write_d->getNext(w_head);

		if(!write_d->oneReadHeadEquals(next_w,consumers)){
			//LOG_MEDIUM("GrowableMultReadQueue "<< Conv<T>::toString() <<" push element " << dbg_next_w);
			(*write_d)[w_head]	= unit[u];
			w_head 				= next_w;

		}else{
			//dbg_next_w = next_w_dbg;
			pushAllocNew(*this,unit);	return;
		}
	}
	write_d->w_head			= w_head;
}


template<class T>
GrowableOneReadQueue<T>::GrowableOneReadQueue(const UserMask& users, const UserID& uid)
	:owner_id(uid),all_users(users){//,dbg_next_w(0),dbg_next_r(0){
	read_d 	= construct(INITIAL_MODULO_MASK);
	write_d = read_d;}

template<class T>
GrowableOneReadQueue<T>::~GrowableOneReadQueue(){
	asrt(read_d != NULL); asrt(write_d != NULL);
	asrt(read_d == write_d); asrt(write_d->r_head==write_d->w_head); //queue has to be emptied before
	data_t::destruct(write_d);}


template<class T>
uint32_t GrowableMultReadQueue<T>::adaptID(const uint32_t id) const {
	asrt(id != owner_id.id);
	return id - ((uint32_t)(id > owner_id.id));
}

template<class T>
GrowableMultReadQueue<T>::GrowableMultReadQueue(const UserMask& users,const UserID& uid,const uint32_t cons)
:owner_id(uid.id),consumers(cons),all_users(users){//,dbg_next_w(0){ //asrt(consumers > 0);
	if(consumers == 0) return;
	write_d	= construct(INITIAL_MODULO_MASK);
	read_ds = new data_t*[consumers];
	//dbg_next_r = new uint32_t[consumers];
	for(uint32_t i=consumers;i--;){read_ds[i]=write_d;}//dbg_next_r[i]=0;}
}

template<class T>
GrowableMultReadQueue<T>::~GrowableMultReadQueue(){
	if(consumers == 0) return;
	asrt(write_d == read_ds[0]); //queue has to be emptied before
	data_t::destruct(write_d);
	//delete[] dbg_next_r;
	delete[] read_ds;}




template<class DTA>
GrowableQueueData<DTA>::GrowableQueueData(const UserMask& all_users,  const uint32_t& mod_mask)
		:super(all_users),modulo_mask(mod_mask),w_head(0),r_head(0){
	LOG3("Allocate new data in GrowableQueueData " << (mod_mask+1)*sizeof(DTA) << " bytes");
	data= new DTA[modulo_mask+1];
}


template<class DTA>
GrowableMultQueueData<DTA>::GrowableMultQueueData(const UserMask& all_users, const uint32_t consumers, const uint32_t& mod_mask)
	:super(all_users),modulo_mask(mod_mask),w_head(0){
	asrt(consumers < 16);
	LOG3("Allocate new data in GrowableMultQueueData " << (mod_mask+1)*sizeof(DTA) << " bytes");
	r_heads = new uint32_t[consumers];
	for(uint32_t i=consumers;i--;){r_heads[i]=0;}
	data	= new DTA[mod_mask+1];
}

template<class DTA>
bool GrowableMultQueueData<DTA>::oneReadHeadEquals(const uint32_t& t,uint32_t cons){
	for(uint i=cons;i;){
		if(r_heads[--i] == t) return true;}
	return false;
}







#endif /* PRODUCERCONSUMERSQUEUE_HPP_ */
