/*
 * SimplyFire.hpp
 *
 *  Created on: 26.02.2011
 *      Author: stephan
 */

#ifndef SIMPLYFIRE_HPP_
#define SIMPLYFIRE_HPP_
#include "Queue.hpp"
#include "Hash.hpp"

struct OccCls{
	uint32_t cli	:31;
	uint32_t pol	:1;
	OccCls(const ClsPointer& cp,const bool polarity):
		cli(cp.idx()),pol(polarity){asrt(cp.idx()<(1<<30));}
	ClsPointer getClsP() const	{return ClsPointer(cli);}
	bool isPosOcc() const	{return pol;}
};


struct ClsBin{
	uint32_t bin		:1;
	uint32_t lit_cls	:31; //literal or clause

	ClsBin(const Literal& l=TRUE_LIT):bin(1),lit_cls(l.idx()){}
	ClsBin(const ClsPointer& c)		:bin(0),lit_cls(c.idx()){}
	bool isBin() const				{return bin;}
	Literal getLit() const			{asrt(isBin());return Literal(lit_cls);}
	ClsPointer getClp() const		{asrt(!isBin());return ClsPointer(lit_cls);}
};




struct SubsumDummy{
	SubsumDummy(const BinVarInfo& v_info, const ClsInfo& c_info,const Watching& watches){}
	void ensureVars(uint32_t sz){}
	void reserveClses(uint32_t sz){}
	void registerBin(const Literal& a, const Literal& b){};
	void registerClause(const ClsPointer& clp, SharedClause* cls){}
	void delOccs(const Variable& var){}
	void cleanWasteAndDeleted(){}

};

struct ClsExtr{
	uint32_t eli_soon	: 1;
	uint32_t mrk		: 1;
	uint32_t sz			: 30;
	uint32_t sig;

	bool mayContainOneOf(const uint32_t& lits_msk){return sig & lits_msk;}
	bool cannotBeSubsumedBy(const uint32_t& sub_sz, const uint32_t& sub_sig){
			return (sz < sub_sz) || (~sig & sub_sig) ;}
};

struct ClsExtrs{
	vector<ClsExtr>	extr;

	ClsExtr& get(const ClsPointer& clp){asrt(clp.idx()<extr.size());
		return extr[clp.idx()];}

	void ensure(const ClsPointer& clp){
		if(clp.idx()>=extr.size()){extr.resize(clp.idx()+1);}}
	void setMrk(const ClsPointer& clp,bool b){get(clp).mrk=b;}
	bool isMrkd(const ClsPointer& clp){return get(clp).mrk;}
};

struct SimplyFire{

	const BinVarInfo& vinf;
	const ClsInfo& cinf;
	const Watching& wtch;
	vector<uint64_t>			lit_space;
	ClsExtrs					cls_extr;
	VecOfLists<OccCls,Variable> occs;
	Queue<ClsPointer,1024>		subsum_q;//clses that may subsume others or be subsumed by others
	Queue<Literal,1024>			subsum_bq;//binaries that may subsume other clauses
	BitVector<Literal>			bin_eq;


	bool bins_udt;

	static uint32_t getHash(const Literal& l){
		return hashBit32(l.toVar().idx());}


	bool dequeueSubLit(Literal& lit){
		if(subsum_bq.pop_front(lit)){	bin_eq.clear(lit);return true;}
		return false;}

	bool dequeueSubCls(ClsPointer& clp){return subsum_q.pop_front(clp);}

	uint32_t clsQueueSize()const{return subsum_q.size();}

	SimplyFire(const BinVarInfo& v_info, const ClsInfo& c_info,const Watching& watches):
		vinf(v_info),cinf(c_info),wtch(watches),occs('O'),bins_udt(false){}



	inline void copyOccsTo(const Variable& var, vector<OccCls>& occs_out);


	inline void registerBin(const Literal& a, const Literal& b);
	inline void registerClause(const ClsPointer& clp, SharedClause* cls);
	Iter<OccCls> getOccs(const Variable& var){return occs.getIter(var);}
	uint32_t occsCount(const Variable& var){return occs.sizeOf(var);}
	void delOccs(const Variable& var){occs.clear(var);}
	inline uint32_t splitFixClses(const Variable& var,vector<ClsBin> split[],const uint32_t& max_lits,const uint32_t& max_fact);
	inline uint32_t splitBinsHlpr(const Literal& a, vector<ClsBin>& add);

	bool binsUpdated(){bool ret=bins_udt;bins_udt=false;return ret;}

	void ensureVars(uint32_t sz){
		occs.enlarge(sz);
		sz *= 2;
		lit_space.resize(sz);
		bin_eq.ensureSize(sz,false);}

	void reserveClses(uint32_t sz){cls_extr.extr.reserve(sz);}
	inline void cleanWasteAndDeleted();
	inline void cleanWasteOfVar(const Variable& var);
	inline uint32_t cleanWasteOfVarCountFix(const Variable& var);
};


void SimplyFire::cleanWasteOfVar(const Variable& var){
	WIter<OccCls> wi = occs.getWIter(var); OccCls oc(ClsPointer(0),false);
	while(wi.getNext(oc)){
		if(cinf.validNoWaste(oc.getClsP())) {//asrt(!cinf.getCls(clp)->isSoonWaste());
			wi.keep(oc);}
	}
	occs.finish(var,wi);
}


uint32_t SimplyFire::cleanWasteOfVarCountFix(const Variable& var){
	uint32_t fix = 0;
	WIter<OccCls> wi = occs.getWIter(var); OccCls oc(ClsPointer(0),false);
	while(wi.getNext(oc)){
		if(!cinf.validNoWaste(oc.getClsP())) {continue;}
		wi.keep(oc);

		ClsExtr& ce = cls_extr.get(oc.getClsP());
		bool bad	= ((bool)ce.eli_soon) || cinf.isLearnt(oc.getClsP());
		fix 		+= (!bad);
	}
	occs.finish(var,wi);
	return fix;
}

void SimplyFire::cleanWasteAndDeleted(){
	for(uint32_t i=occs.listCount();i--;){
		cleanWasteOfVar(Variable(i));
	}
}


void 	SimplyFire::copyOccsTo(const Variable& var, vector<OccCls>& occs_out){
	WIter<OccCls> wi = occs.getWIter(var); OccCls oc(ClsPointer(0),false);
	while(wi.getNext(oc)){
		if(cinf.validNoWaste(oc.getClsP())){
			occs_out.push_back(oc);
			wi.keep(oc);
		}
	}
	occs.finish(var,wi);
}


void SimplyFire::registerBin(const Literal& a,const Literal& b){
	bins_udt = true;
	if(bin_eq.isSet(a) || bin_eq.isSet(b)){ return;}

	bin_eq		.setFlag(a);
	subsum_bq	.push_back(a);
}

void SimplyFire::registerClause(const ClsPointer& clp, SharedClause* cls){
	cls_extr.ensure(clp);

	subsum_q		.push_back(clp);
	ClsExtr& extra 	= cls_extr.get(clp);
	uint32_t sz=0; uint32_t sig=0; bool soon_eli = 0;
	FOR_LITS_STRT(cls,lit);
		++sz;
		sig |= getHash(lit);
		occs.add(lit.toVar(),OccCls(clp,lit.isPos()));
		soon_eli |= vinf.litElimSoon(lit);
	FOR_LITS_END;

	//extra.lrnt	= cls->isLearnt();
	extra.eli_soon 	= soon_eli;
	extra.mrk		= 0;
	extra.sz		= sz;
	extra.sig		= sig;
}



//split has size 2 -> one vector for negative occurrences, one for positive
uint32_t SimplyFire::splitFixClses(const Variable& var,vector<ClsBin> split[],const uint32_t& max_lits, const uint32_t& max_fact){
	asrt(split[0].empty()); asrt(split[1].empty());
	LOG2("Split clauses of var " << var.toDIMACS());

	OccCls oc(ClsPointer(0),false);
	uint32_t lits	= 0;
	Iter<OccCls> oi = occs.getIter(var);

	while(oi.getNext(oc)){ //do not delete waste anymore
		ClsPointer clp = oc.getClsP(); asrt(cinf.validPtr(clp));
		if(cinf.isLearnt(clp)) continue;

		ClsExtr& ec 	= cls_extr.get(clp);
		if(ec.eli_soon){		continue;}

		lits			+= ec.sz;

		if(lits > max_lits) { LOG2("Early break of splitFixClses max was " << max_lits << " we have " << lits);
			return max32;}

		split[oc.isPosOcc()]	.push_back(ClsBin(clp));

		if(split[oc.isPosOcc()].size() > max_fact){
			LOG2("Early break of splitFixClses max fact was " << max_fact);
			return max32;}
	}

	lits 	+= 	splitBinsHlpr(var.toNegLit(),split[0])
			+	splitBinsHlpr(var.toPosLit(),split[1]);

	return lits;
}


uint32_t SimplyFire::splitBinsHlpr(const Literal& a, vector<ClsBin>& add){
	Iter<Literal> it = wtch.getBinsOf(a); Literal lit;
	const uint32_t s = add.size();
	while(it.getNext(lit)){ if(!vinf.litElimSoon(lit)){add.push_back(ClsBin(lit));}}
	return add.size() - s; //only count literals not double
}



#endif /* SIMPLYFIRE_HPP_ */
