/*
 * SolverFactory.hpp
 *
 *  Created on: 13.08.2009
 *      Author: stephan
 */

#ifndef SOLVERFACTORY_HPP_
#define SOLVERFACTORY_HPP_
#define solverfactory_hpp_version	  "$Revision: 1.8 $"
const char*  solverfactory_cpp_version();

#include "GlobalSettings.hpp"
#include "Satistics.hpp"
#include "BCPEngine.hpp"
#include "SolverLeader.hpp"
#include "SolverTrooper.hpp"
#include "IFStream.hpp"
#include "Decision.hpp"
#include "SiblingInfo.hpp"
#include "SimplyFire.hpp"
#include "ElimClses.hpp"
#include "DMRPTrooper.hpp"

using std::vector;



template<class SOLVERS>
bool parse(IFStream& in, SOLVERS& solvers) {
    vector<Literal> prsd;
    int32_t 		plit;

    while(in.hasNext()){
        if(in.currEquals('c')) {	in.skipLine();}
        else if(in.currEquals('p')){
            if (in == "p cnf"){
                solvers.reserveForVariables(in.getNextInt()+1);
                solvers.reserveForClauses(in.getNextInt());
                in.skipLine();
            }else{
            	std::cerr << "c PARSE ERROR" << std::endl ;
            	exit(EXIT_FAILURE); }
        }else{
            while((plit = in.getNextInt()) != 0){
            	prsd.push_back(Literal::DIMACS2Literal(plit));
            }
            if(!solvers.addClause(prsd,false)){
                solvers.invalidateAll();return false;}

            prsd.clear();
        }
    }
    return true;
}

template<class BCPT>
class SolverLeader;

class SingleSolver{
	uint32_t 	printed;
	ElimClses 	elim;

	typedef VarAct								decision_t;
	typedef BcpEngine<decision_t,LeaderSiblingInfo,SimplyFire>
												bcp_t;

	typedef SolverLeader<	bcp_t> 				solver_t;

	solver_t*  					solver;


public:
	SingleSolver(GlobalSettings& sets);
	void 	initSharing()const {};
	bool 	parseDIMACS(IFStream& in){return parse(in, *this);}
	bool 	addClause(vector<Literal>& literals,const bool lrnt);
	bool	canPrintFinal(){return exchange32(&printed,0,1);}
	int		printResult(const UserID& uid,const string& file,TrBool res);
	void	printDIMACS(const UserID& uid,const std::string& file,TrBool res);
	void 	printCSV(const UserID& uid,const std::string& file_name, TrBool result);
	void 	printStatsConsole(const UserID& uid, TrBool result);
	void 	checkOut(const UserID& uid);
	TrBool	invokeSolver(const UserID& uid);
	void 	stopPrinting();
	void 	stopAllASAP();
	void 	reserveForVariables(uint32_t s);
	void 	reserveForClauses(uint32_t s);
	void 	invalidateAll();
};



struct ExplicitSolver{

	typedef SolverLeader<BcpEngine<VarAct,LeaderSiblingInfo,SimplyFire> >	lead_t;
	typedef SolverTrooper<BcpEngine<VarAct,TrooperSiblingInfo,SubsumDummy> >	troop_t1;
	typedef SolverTrooper<BcpEngine<VarAct,TrooperSiblingInfo,SubsumDummy> >	troop_t2;
	typedef DMRPTrooper															troop_t3;
	//typedef SolverTrooper<BcpEngine<VarAct,TrooperSiblingInfo,SubsumDummy> >	troop_t3;

	typedef SolverTrooper<BcpEngine<VarAct,TrooperSiblingInfo,SubsumDummy> >	troop_t4;
	//typedef SolverTrooper<BcpEngine<LitAct,TrooperSiblingInfo,SubsumDummy> >	troop_t5;
	//typedef DMRPTrooper															troop_t6;
	typedef SolverTrooper<BcpEngine<LitAct,TrooperSiblingInfo,SubsumDummy> >	troop_t5;
	typedef SolverTrooper<BcpEngine<VarAct,TrooperSiblingInfo,SubsumDummy> >	troop_t6;
	typedef SolverTrooper<BcpEngine<VarAct,TrooperSiblingInfo,SubsumDummy> >	troop_t7;

protected:

	ElimClses						elim;
	UserMask						running;
	uint8_t 						printed;
	const uint8_t					threads;

	lead_t*  						leader;
	troop_t1*						trooper1;
	troop_t2*						trooper2;
	troop_t3*						trooper3;
	troop_t4*						trooper4;
	troop_t5*						trooper5;
	troop_t6*						trooper6;
	troop_t7*						trooper7;

	template<class T>
	inline void linkOthers(T* trooper);


public:
	void initTypes(GlobalSettings& sets);
	ExplicitSolver(uint32_t thds):running(0),printed(0),threads(thds){};
	void 	initSharing();
	bool 	parseDIMACS(IFStream& in){return parse(in, *this);}
	bool 	addClause(vector<Literal>& literals,const bool lrnt);
	bool	canPrintFinal(){return exchange8(&printed,0,1);}
	int		printResult(const UserID& uid,const string& file,TrBool res);
	//void	printDIMACS(const UserID& uid,const string& file,TrBool res);
	void 	printCSV(const UserID& uid,const std::string& file_name, TrBool result);
	void 	printStatsConsole(const UserID& uid, TrBool result);
	void 	checkOut(const UserID& uid);
	TrBool	invokeSolver(const UserID& uid);
	void 	stopPrinting();
	void 	stopAllASAP();
	void 	reserveForVariables(uint32_t s);
	void 	reserveForClauses(uint32_t s);
	void 	invalidateAll();
};



/*
struct MultSolver{

	//typedef Conflicts2MaxLearnts 								confl_t;
	typedef VarAct						 						decision_t;

	typedef BcpEngine<decision_t,LeaderSiblingInfo,SimplyFire>
																bcp_indept_t;

	typedef SolverLeader<bcp_indept_t> 							lead_solver_t;

	typedef BcpEngine<decision_t,TrooperSiblingInfo,SubsumDummy>
																bcp_dept_t;

	typedef TrooperSiblingInfo									troop_sinfo_t;
	typedef SolverTrooper<bcp_dept_t> 							troop_solver_t;
private:

	ElimClses						elim;
	uint32_t 						printed;
	UserMask						running;
	lead_solver_t*  				lead_solver;
	vector<troop_solver_t* > 		troop_solvers;

public:
	MultSolver(GlobalSettings& sets);
	bool 	initSharing();
	bool 	parseDIMACS(IFStream& in){return parse(in, *this) && initSharing();}
	bool 	addClause(vector<Literal>& literals,const bool lrnt);
	bool	canPrintFinal(){return exchange32(&printed,0,1);}
	int		printResult(const UserID& uid,const string& file,TrBool res);
	void	printDIMACS(const UserID& uid,const string& file,TrBool res);
	void 	printCSV(const UserID& uid,const std::string& file_name, TrBool result);
	void 	printStatsConsole(const UserID& uid, TrBool result);
	void 	checkOut(const UserID& uid);
	TrBool	invokeSolver(const UserID& uid);
	void 	stopPrinting();
	void 	stopAllASAP();
	void 	reserveForVariables(uint32_t s);
	void 	reserveForClauses(uint32_t s);
	void 	invalidateAll();
};
*/


/*struct DebugSolver: public ExplicitSolver{
	void initTypes(GlobalSettings& sets);
	DebugSolver(uint32_t thds): ExplicitSolver(thds){}

};*/


#define link_if(t,id) if(trooper ## t != NULL && id != t){tsinf->addTrooper(trooper ## t->sstate.s_info);}

template<class T>
void ExplicitSolver::linkOthers(T* trooper){
	if(trooper==NULL) return; asrt(threads <=8);
	TrooperSiblingInfo* tsinf = trooper->sstate.s_info;

	leader->sstate.s_info	->addTrooper(tsinf);
	tsinf					->addLeader(leader->sstate.s_info);
	const uint32_t id	= tsinf->user_id.id;

	link_if(1,id)
	link_if(2,id)
	link_if(3,id)
	link_if(4,id)
	link_if(5,id)
	link_if(6,id)
	link_if(7,id)
}


#endif /* SOLVERFACTORY_HPP_ */

