#ifndef WATCHING_HPP_
#define WATCHING_HPP_
#define watching_hpp_version "$Revision: 1.7 $"

#include "GlobalSettings.hpp"
#include <vector>
#include "VecOfLists.hpp"
#include "ClauseInfos.hpp"
#include "VariableInfos.hpp"



struct WtchEle{
	Literal		other;
	ClsPointer 	clp;
	WtchEle(const Literal& l, const ClsPointer& p):other(l),clp(p){}
	WtchEle(){}
};


class Watching {
	VecOfLists<Literal,Literal>			bins;
	VecOfLists<WtchEle, Literal>		wtchd;

public:
	Watching():bins('B'),wtchd('W'){}
	void 	createClone(const Watching& clone){bins.initWith(clone.bins); wtchd.initWith(clone.wtchd);}
	uint32_t 		litsCount() const{return bins.listCount();}

	Iter<Literal> getBinsOf(const Literal& a)const{return bins.getIter(a);}
	WIter<Literal>	getWBinsOf(const Literal& a){return bins.getWIter(a);}
	void			finishBins(const Literal& a, WIter<Literal> wi){bins.finish(a,wi);}
	Range<Literal>  getBinsRangeOf(const Literal& a){return bins.getRange(a);}
	bool			getNextBin(Range<Literal>& r, Literal& out){return bins.getNext(r,out);}
	void			replaceNextBin(Range<Literal>& r, const Literal& n){bins.writeNext(r,n);}
	void			finishBinReplace(const Literal& a, Range<Literal>& r){bins.finishWrite(a,r);}
	void			clearBinsPartOf(const Literal& a){bins.clear(a); LOG2("Cleared all bins of " << a.toDIMACS());}
	bool			noBinsPart(const Literal& a)const{return bins.empty(a);}
	uint32_t		nbrBins(const Literal& a)const{return bins.sizeOf(a);}

	Iter<Literal> getImpls(const Literal& a){return bins.getIter(a.flipped());}

	WIter<WtchEle> 	getWtchOfW(const Literal& l)	{return wtchd.getWIter(l);}
	Iter<WtchEle> 	getWtchOf(const Literal& l)	const {return wtchd.getIter(l);}
	Literal 		getLitWithSmallerList(const Variable& v)	{return wtchd.sizeOf(v.toPosLit()) < wtchd.sizeOf(v.toNegLit()) ? v.toPosLit() : v.toNegLit();}
	bool isEmpty(const Literal& lit) const		{return wtchd.sizeOf(lit)==0;}
	void finish(const Literal& lit, WIter<WtchEle> it){wtchd.finish(lit,it);}
	void finish(const Literal& lit, WIter<Literal> it){bins.finish(lit,it);}
	void keepRest(const Literal& lit, WIter<WtchEle>& iter){
			while(iter.hasNext()){iter.keepNext();}
			wtchd.finish(lit,iter);}
	void clearOf(const Literal& l){wtchd.clear(l);bins.clear(l);}

	void enlarge(const uint32_t& sz){wtchd.enlarge(sz);bins.enlarge(sz);}

	inline void	removeBinPart(const Literal& of, const Literal& who);
	inline void addClsesOf(const Literal& lit, vector<ClsPointer>& out);
	inline void addCls(const Literal& lit, const ClsPointer& clp, const Literal& cached);
	inline void	addBinCls(const Literal& a, const Literal& b);
	inline void	addBinClsIf(const Literal& a, const Literal& b);
	inline bool containsBin(Literal a, Literal b)const;
	//inline bool  containsImpl(const Literal& from, const Literal& to)const;

	inline void removeVar(const Variable& die, vector<Literal>& update_bins);
	inline void addBins(const Literal& die, const Literal& survive,vector<Literal>& out, vector<Literal>& units);
	//inline bool replaceVar(const Literal& die, const Literal& survive,vector<Literal>& update_bins,	vector<ClsPointer>& update_cls);

	/*inline Literal findLitRepl(Literal a) const;
	inline bool replaceVar(const Literal& die, const Literal& survive,vector<Literal>& update_impls,
			vector<ClsPointer>& update_cls);

	inline void updateReplOfVar(const Variable& var);*/

	inline bool asrtContainsBinOneDir(const Literal& a, const Literal& b) const;
	inline bool asrtBinsBothDirs(const BinVarInfo& v_info)const;
	inline bool asrtPathExists(const Literal& from, const Literal& to, const BinVarInfo& v)const;
};


void Watching::addClsesOf(const Literal& lit, vector<ClsPointer>& out){
	Iter<WtchEle> iter 	= wtchd.getIter(lit);
	WtchEle we;
	while(iter.getNext(we)){out.push_back(we.clp);}
}


void Watching::removeBinPart(const Literal& of, const Literal& who){
	LOG2("Binary in " << of.toDIMACS() << " let die " << who.toDIMACS());
	WIter<Literal> wi = bins.getWIter(of); Literal cand; asrt(of != who);
	while(wi.getNext(cand)){
		if(cand==who){continue;}
		wi.keep(cand);
	}
	bins.finish(of,wi);
}




void Watching::removeVar(const Variable& die, vector<Literal>& update_bins){

	update_bins.push_back(die.toNegLit());
	update_bins.push_back(die.toPosLit());
	Literal ob;
	Iter<Literal> bi1 = bins.getIter(die.toPosLit());
	while(bi1.getNext(ob)){update_bins.push_back(ob);}

	Iter<Literal> bi2 = bins.getIter(die.toNegLit());
	while(bi2.getNext(ob)){update_bins.push_back(ob);}
}



void Watching::addBins(const Literal& die, const Literal& survive,vector<Literal>& out, vector<Literal>& units){
	Iter<Literal> iter = bins.getIter(die); Literal ol; const Literal sf = survive.flipped();
	while(iter.getNext(ol)){
		if(ol==sf) continue;
		if(ol==survive){
			LOG1("Unit in lit replace: " << survive.toDIMACS());
			units.push_back(survive);
		}else{
			LOG3("Update impls of lit " << ol.toDIMACS());
			bins.add(survive,ol);
		}
		out.push_back(ol);
	}
}


bool  Watching::containsBin(Literal a, Literal b)const{
	Literal t;
	if(bins.sizeOf(b) < bins.sizeOf(a)){t=b;b=a;a=t;}
	Iter<Literal> it = bins.getIter(a);
	while(it.getNext(t)){
		if(t==b) return true;
	}
	return false;
}


void Watching::addBinCls(const Literal& a, const Literal& b){
	asrt(a != b); asrt(a.flipped() != b);
	bins.add(a,b);
	bins.add(b,a);
}

void Watching::addBinClsIf(const Literal& a, const Literal& b){
	asrt(a != b); asrt(a.flipped() != b);
	if(!containsBin(a,b)){addBinCls(a,b);}
}


void  Watching::addCls(const Literal& lit, const ClsPointer& clp, const Literal& cached){
	wtchd.add(lit,WtchEle(cached,clp));
	LOG3("add clause to " << lit.toDIMACS() << " ptr " << clp.idx() << " cached " << cached.toDIMACS());
}


bool Watching::asrtContainsBinOneDir(const Literal& a, const Literal& b) const{
	Iter<Literal> it = bins.getIter(a); Literal o;
	while(it.getNext(o)){
		if(o==b) return true;
	}
	return false;
}

bool Watching::asrtBinsBothDirs(const BinVarInfo& v_info)const{
	bool ok = true;
	for(uint32_t i=wtchd.listCount();i--;){

		Literal l(i);
		if(v_info.litFix(l) || v_info.varRepl(l.toVar())) continue;

		Literal o;
		Iter<Literal> iter = bins.getIter(l);
		while(iter.getNext(o)){
			if(v_info.litFix(o)) continue;

			if(v_info.varRepl(o.toVar())){ ok=false;
				ASRT_PRNT("Normal lit " << l.toDIMACS() << " points to " << o.toDIMACS() << " but this is replaced"<<endl);
			}

			if(v_info.varElim(o.toVar())){ ok=false;
				ASRT_PRNT("Normal lit " << l.toDIMACS() << " points to " << o.toDIMACS() << " but this is elim"<<endl);
			}

			if(!asrtContainsBinOneDir(o,l)){
				ASRT_PRNT("Normal lit " << l.toDIMACS() << " points to " << o.toDIMACS() << " but no other dir"<<endl);
				ok=false;
			}
		}
	}
	//ASRT_PRNT("Finished asrtImplsBothDirs"<<endl);
	return ok;
}

inline bool Watching::asrtPathExists(const Literal& from, const Literal& to, const BinVarInfo& v)const{
	LOG3("Check for path " << from.toDIMACS() << " to " << to.toDIMACS());
	if(v.litFix(from) || v.litFix(to)) return true;

	if(v.litElim(from) || v.litRepl(from)){
		ASRT_PRNT("asrtPathExists with replaced lits as arguments" << endl);return false;}
	if(to == from){
		//ASRT_PRNT("path exists from " << from.toDIMACS() << " to " <<to.toDIMACS()<< endl);
		return true;
	}

	vector<Literal> queue;
	queue.push_back(from); uint32_t qhead = 0;
	Literal target;
	BitVector<Literal> flgs; flgs.ensureSize(wtchd.listCount(),false);

	while(qhead != queue.size()){
		Literal src 		= queue[qhead++];
		Iter<Literal> ii 	= bins.getIter(src.flipped());

		while(ii.getNext(target)){

			if(flgs.isSet(target)) continue;

			if(target == to){
				LOG3("Path exsits: " << from.toDIMACS() << " to " << to.toDIMACS());
				//ASRT_PRNT("path exists from " << from.toDIMACS() << " to " <<to.toDIMACS()<< endl);
				return true;}

			flgs	.setFlag(target);
			queue	.push_back(target);
		}
	}
	ASRT_PRNT("path not exists from " << from.toDIMACS() << " to " <<to.toDIMACS()<< endl);

	return false;
}


/*

void Watching::getImplsFlpd(Literal lit, vector<Literal>& out){
	ImplIter iter = getImplIter(lit);
	while(iter.getNext(lit)){	out.push_back(lit.flipped());}
}






void Watching::updateReplOfVar(const Variable& var){
	asrt(isVarRepl(var));
	Literal n 	= var.toNegLit();
	Literal nr	= findLitRepl(n); asrt(!isLitRepl(nr));

	heads[n.idx()].head 			= nr.idx();
	heads[n.flipped().idx()].head 	= nr.flipped().idx();
}



bool  Watching::containsImpl(const Literal& from, const Literal& to) const{
	ImplIter iter = getImplIter(from); Literal lit;
	while(iter.getNext(lit)){
		if(lit == to) return true;}
	return false;
}



void  Watching::addOneImpl(const Literal& lit, const Literal& b){
	uint32_t blk 			= ensureGetBeg(lit,1); asrt(lit != b);
	blks[binEnd(blk)++]	= b.idx();
}


void  Watching::addCls(const Literal& lit, const ClsPointer& clp, const Literal& cached){
	asrt(!isLitRepl(lit));
	uint32_t blk 			= ensureGetBeg(lit,2);
	uint32_t lbeg			= longBeg(blk);
	blks[--lbeg]			= cached.idx();
	blks[--lbeg]			= clp.idx(); asrt(lbeg >= binEnd(blk));
	longBeg(blk)			= lbeg;
	LOG3("add clause to " << lit.toDIMACS() << " ptr " << clp.idx() << " cached " << cached.toDIMACS());
}




*/

#endif /* WATCHING_HPP_ */
